package com.devsoap.json;

import jakarta.json.JsonArray;
import jakarta.json.JsonNumber;
import jakarta.json.JsonString;
import jakarta.json.JsonValue;
import org.assertj.core.api.*;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.regex.Pattern;

import static com.devsoap.json.JsonAssert.typedJsonArray;

public class JsonValueAssert<VALUE_TYPE> extends AbstractAssert<JsonValueAssert<VALUE_TYPE>, VALUE_TYPE> {

    private final String fieldName;
    private final JsonValue value;

    protected JsonValueAssert(String fieldName, JsonValue value, VALUE_TYPE fieldValue) {
        super(fieldValue, JsonValueAssert.class);
        this.fieldName = fieldName;
        this.value = value;
    }

    public JsonValueAssert<?> field(String fieldName) {
        if (value.getValueType() != JsonValue.ValueType.OBJECT) {
            failWithActualExpectedAndMessage(value.getValueType(), JsonValue.ValueType.OBJECT,
                    "Only objects can have fields");
        }
        return JsonAssert.typedJsonValue(fieldName, value.asJsonObject().get(fieldName))
                .orElseGet(() -> this.withFailMessage("Field %s not found", fieldName));
    }

    public JsonValueAssert<VALUE_TYPE> field(String fieldName, Class<VALUE_TYPE> type) {
        if (value.getValueType() != JsonValue.ValueType.OBJECT) {
            failWithActualExpectedAndMessage(value.getValueType(), JsonValue.ValueType.OBJECT,
                    "Only objects can have fields");
        }
        return (JsonValueAssert<VALUE_TYPE>) JsonAssert.typedJsonValue(fieldName, value.asJsonObject().get(fieldName))
                .orElseGet(() -> this.withFailMessage("Field %s not found", fieldName));
    }

    public JsonArrayAssert<?> array(String fieldName) {
        if (value.getValueType() != JsonValue.ValueType.OBJECT) {
            failWithActualExpectedAndMessage(value.getValueType(), JsonValue.ValueType.OBJECT,
                    "Only objects can have fields");
        }

        JsonValue jsonValue = value.asJsonObject().get(fieldName);
        if (jsonValue.getValueType() != JsonValue.ValueType.ARRAY) {
            failWithActualExpectedAndMessage(value.getValueType(), JsonValue.ValueType.OBJECT,
                    "Field must be an array");
        }

        return typedJsonArray(fieldName, jsonValue);
    }

    public <ITEM> ListAssert<ITEM> array(String fieldName, Class<ITEM> itemClass) {
        if (value.getValueType() != JsonValue.ValueType.OBJECT) {
            failWithActualExpectedAndMessage(value.getValueType(), JsonValue.ValueType.OBJECT,
                    "Only objects can have fields");
        }

        JsonValue jsonValue = value.asJsonObject().get(fieldName);
        if (jsonValue.getValueType() != JsonValue.ValueType.ARRAY) {
            failWithActualExpectedAndMessage(value.getValueType(), JsonValue.ValueType.OBJECT,
                    "Field must be an array");
        }

        JsonArray array = (JsonArray) jsonValue;
        return new ListAssert<>(array.getValuesAs(value -> (ITEM) switch (value.getValueType()) {
            case ARRAY -> throw failure("Cannot convert ARRAY to %s", itemClass);
            case OBJECT -> throw failure("Cannot convert OBJECT to %s", itemClass);
            case STRING -> ((JsonString) value).toString();
            case NUMBER -> ((JsonNumber) value).intValue();
            case TRUE -> Boolean.TRUE;
            case FALSE -> Boolean.FALSE;
            case NULL -> null;
        }));
    }

    public JsonValueAssert<?> path(String... fields) {
        JsonValueAssert<?> currentField = this;
        for (String field : fields) {
            currentField = currentField.field(field);
        }
        return currentField;
    }

    public IntegerAssert asInteger() {
        if (actual instanceof Integer) {
            return new IntegerAssert((Integer) actual);
        }
        return new IntegerAssert(Integer.parseInt(actual.toString()));
    }

    public BigDecimalAssert asBigDecimal() {
        if (actual instanceof BigDecimal) {
            return new BigDecimalAssert((BigDecimal) actual);
        }
        return new BigDecimalAssert(BigDecimal.valueOf(Double.parseDouble(actual.toString())));
    }

    public BigIntegerAssert asBigInteger() {
        if (actual instanceof BigInteger) {
            return new BigIntegerAssert((BigInteger) actual);
        }
        return new BigIntegerAssert(BigInteger.valueOf(Integer.parseInt(actual.toString())));
    }

    public DoubleAssert asDouble() {
        if (actual instanceof Double) {
            return new DoubleAssert((Double) actual);
        }
        return new DoubleAssert(Double.parseDouble(actual.toString()));
    }

    public AbstractLongAssert<?> asLong() {
        if (actual instanceof Long) {
            return new LongAssert((Long) actual);
        }
        return new LongAssert(Long.parseLong(actual.toString()));
    }

    public JsonArrayAssert<?> asArray() {
        if (value.getValueType() == JsonValue.ValueType.ARRAY) {
            return new JsonArrayAssert<>(fieldName, (JsonArray) value);
        }
        return new JsonArrayAssert<>(fieldName, value);
    }

    public JsonValueAssert<?> matches(String regex) {
        return matches(valueType -> Pattern.matches(regex, valueType.toString()));
    }

    public JsonValueAssert<?> notMatches(String regex) {
        return matches(valueType -> !Pattern.matches(regex, valueType.toString()));
    }

    @Override
    public JsonValueAssert<VALUE_TYPE> isEqualTo(Object expected) {
        return super.isEqualTo(expected);
    }
}